#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

#include "debug.h"
#include "kprintf.h"
#include "memory.h"
#include "string.h"
#include "utils.h"

static i16 find_next_block(size_t nb_pages, u16 *pd_index_ptr,
                               u32 **page_table_ptr)
{
	for (*pd_index_ptr = 1; *pd_index_ptr < 768; (*pd_index_ptr)++) {
		if (page_directory[(*pd_index_ptr)] == 0x02) {
			if (add_page_table(*pd_index_ptr) < 0)
				return -2;
		}
		*page_table_ptr =
		    (u32 *)GET_PAGE_ADDR(0, *pd_index_ptr + PT_START);
		for (u16 i = 0; i + nb_pages - 1 < PT_SIZE; i++) {
			u16 j;
			for (j = 0; (*page_table_ptr)[i + j] >> 12 == i + j &&
			            j < nb_pages;
			     j++)
				;
			if (j == nb_pages)
				return i;
			i += j;
		}
	}
	return -1;
}

i8 add_single_page(void *frame)
{
	u16 pd_index;
	u32 *page_table;
	const i16 i = find_next_block(1, &pd_index, &page_table);

	if (i < 0) {
		kprintf(KERN_CRIT "impossible to add page to page directory\n");
		return -1;
	}
	page_table[i] = ((u32)frame & PAGE_MASK) | INIT_FLAGS;
	return 0;
}

void *alloc_pages(size_t size, void **phys_addr)
{
	const u32 nb_pages = CEIL(size, PAGE_SIZE);
	u16 pd_index;
	u32 *page_table;
	const i16 index = find_next_block(nb_pages, &pd_index, &page_table);

	if (index < 0) {
		kprintf(KERN_CRIT "%d: Not enough pages (max: %d)\n", index,
		        PT_SIZE);
		return NULL;
	}
	for (size_t i = index; i - (size_t)index < nb_pages; i++) {
		void *frame = alloc_frame();
		if (!frame) {
			for (size_t j = index; j < i; j++)
				free_frame((void *)(page_table[j] & PAGE_MASK));
			return NULL;
		}
		if (phys_addr)
			*phys_addr = frame;
		page_table[i] = ((u32)frame & PAGE_MASK) | INIT_FLAGS;
	}
	memset((void *)GET_PAGE_ADDR(pd_index, index), 0, nb_pages * PAGE_SIZE);
	return (void *)GET_PAGE_ADDR(pd_index, index);
}

int free_pages(void *page_ptr, size_t size)
{
	const u32 page_addr = (u32)page_ptr;
	const u32 nb_pages = CEIL(size, PAGE_SIZE);
	const u32 page_index = page_addr / PAGE_SIZE;
	const u32 pd_index = page_index / PD_SIZE;
	const u32 pt_index = page_index % PD_SIZE;

	if ((u32)pd_index > 0x300) {
		kprintf(KERN_WARNING "Address out of range\n");
		return -1;
	} else if (page_addr % PAGE_SIZE) {
		kprintf(KERN_WARNING "Invalid address\n");
		return -1;
	} else if (pt_index + nb_pages > PT_SIZE) {
		kprintf(KERN_WARNING "Invalid number of frames\n");
		return -1;
	}
	u32 *page_table =
	    (u32 *)GET_PAGE_ADDR(0, PT_START + pd_index);
	for (u16 i = pt_index; i < pt_index + nb_pages; i++) {
		if (page_table[i] >> 12 == i) {
			kprintf(KERN_WARNING "Page already free\n");
			return -2;
		}
		free_frame((void *)(page_table[i] & PAGE_MASK));
		page_table[i] = i << 12;
	}

	return 0;
}
